/****************************************************************************/
/*                         TUOX kernel: modul KLIB88.C                      */
/****************************************************************************/

/**********************************************************
 Soubor s pomocnymi assemblerovymi rutinami pro kernel
*************************************************************/
//	k_phys_copy (dword src, dword dest, word count)
//		- kopie dat odkudkoli kamkoli v ramci cele pameti
//		(adresy jsou 32-bitove linearni)
//	k_cp_mess (src,src_addr,dest_addr)
//		- kopie message od odesilatele k prijemci + nastavi v nem sendera
//	k_port_out (port, value)
//		- zapis hodnoty na port
//	k_port_in (port)
//		- cteni hodnoty z portu
//	k_lock ()
//		- zakaz preruseni, uschova puvodni hodnotu (povoleno/zakazano)
//		do globalni promenne
//		(musela byt prejmenovana kvuli vyhrazene instrukci LOCK)
//	k_restore ()
//		- obnoveni predchoziho stavu interrupts (z glob. promenne)
//	k_unlock ()
//		- povoleni preruseni
//	k_vid_copy (?)
//		- zapis dat do video RAM (s cekanim na retrace (?) )
//	k_reboot ()
//		- po CTRL-ALT-DEL (obnovi vektory preruseni, vola int 19h)
//		??? csv, cret - uvidime, jestli to bude BC++ potrebovat



/* reboot() - tab. vektoru preruseni - linker-najit */

#include "COMMON/type.h"
#define IFMASK	0x200			// Mask for Interrupt Flag
#define VID_SEG	0xb800
#define MK_FP( seg,ofs )( (void _seg * )( seg ) +( void near * )( ofs ))

unsigned int IF = 1;  // Interrupt flag IF 0 nebo 1

//=============================================================================
// k_phys_copy (dword src, dword dest, word count)
// - kopie dat odkudkoli kamkoli v ramci cele pameti
//=============================================================================

// P.G.: Prepsal jsem to do C, protoze puvodni assembler neumel kopirovani
// prekrytych bloku a v assembleru se mi to psat nechtelo
// VELMI PROVIZORNI A VELMI POMALE

void k_phys_copy(DWord src, DWord dest, Word count) {
 Word sseg,sofs,dseg,dofs;
 char huge* s, huge* d;

 sseg=(src & 0xf0000)>>4;
 sofs=src & 0xffff;
 dseg=(dest & 0xf0000)>>4;
 dofs=dest & 0xffff;
 s=MK_FP(sseg,sofs);
 d=MK_FP(dseg,dofs);


 if (dest==src) return;

 if (dest>src)
  // odshora z dst
   while (count--)
    *(d+count)=*(s+count);
  else
   // odspoda v dst
   while (count--) *d++=*s++;
}

/*
void k_phys_copy(DWord src, DWord dest, Word count)
{
  // ? schovani registru ? zrejme nebude nutne
  // kontrolovat, je-li nad 1 MB? - nekdy dodelat
  // kopirovani bude fungovat pokud se bloky nebudou prekryvat
  // a count<=65520 (65535-15)

  asm push ds
  asm mov bx, 16
  asm mov dx, word ptr [src+2]  //dx=vyssi word ze source
  asm mov ax, word ptr [src]    //ax=nizsi word ze source
  asm div bx                    //ax=dx~ax/bx, zbytek v dx
  // v ax bude segment, v dx bude offset <0..15>
  asm mov si, dx
  asm mov ds, ax                //source segment:offset ... DS:SI

  asm mov dx, word ptr [dest+2]
  asm mov ax, word ptr [dest]
  asm div bx                    //ax=dx~ax/bx, zbytek v dx
  asm mov di, dx
  asm mov es, ax                //destination segment:offset ... ES:DI

  asm mov cx, word ptr [count]  //cx delka bloku do 65520
  asm cld                       //DF=0 => offsety se inkrementuji
  asm rep movsb                 //kopirovani z DS:SI do ES:DI po 1 Byte
  asm pop ds
}
*/
//=============================================================================
// k_cp_mess (src,src_addr,dest_addr)
// - kopie message od odesilatele k prijemci + nastavi v nem sendera
//=============================================================================

void k_cp_mess(Word src, DWord src_addr, DWord dest_addr)
{
	k_phys_copy (src_addr, dest_addr, sizeof (message));
	asm mov ax, word ptr [dest_addr]
	asm mov bx, word ptr [dest_addr+2]
	asm mov si, ax
	asm ror bx, 1
	asm ror bx, 1
	asm ror bx, 1
	asm ror bx, 1	  							//posun o 4 bity
	asm mov es, bx								//dest. segment(DS):offset(SI)
	asm mov ax, src
	asm mov word ptr es:[si], ax     //preplacnu sendera
}

//=============================================================================
// k_port_out (word port, byte value) - zapis hodnoty na port
//=============================================================================

void k_port_out(Word port, Byte value)
{
	asm mov	dx, port
	asm mov	al, value
	asm out	dx, al
}

//=============================================================================
// k_port_in (word port) - cteni hodnoty z portu (vraci v AL)
//=============================================================================
#pragma warn -rvl
Byte k_port_in (Word port)
{
	asm mov	dx, port
	asm in	ax, dx
}
#pragma warn +rvl

//=============================================================================
// k_lock () - zakaz preruseni, uschova puvodni hodnotu (povoleno/zakazano)
//       do globalni promenne
//=============================================================================
void k_lock (void)
{
	asm push ax
	asm pushf
	asm pop ax		       // Load Flags to AX
	asm and	ax, IFMASK	       // Mask Interrupt Flag
	asm jz h1		       // ZF=1 IF=0
	asm mov IF,1
	asm jmp h2
h1:
	asm mov IF,0
h2:
	asm cli
	asm pop	ax
}

//=============================================================================
// k_restore () - obnoveni predchoziho stavu interrupts (z glob. promenne)
//=============================================================================

void k_restore (void)
{
	asm push ax
	asm mov	ax, IF		// Move value from global variable LOCKED to AH
	asm or ax, ax	    		// Test for nonzero
	asm jnz	k_r1		    	// Nonzero
	asm cli			        	// locked = 0  zakaz preruseni
	asm jmp	k_r2	    		// Jump to end of routine
k_r1:
	asm sti				       // locked <>0 dovol preruseni
k_r2:
	asm pop	ax
}



//=============================================================================
// k_unlock () - povoleni preruseni
//=============================================================================

void k_unlock (void)
{
	asm sti
	asm mov IF, 1		//1 - preruseni povoleno
}

//=============================================================================
// k_vid_copy - zapis dat do video RAM
//=============================================================================
/* NEZKONTROLOVANO !!! */

void k_vid_copy (Word src_seg, Word src_off, Word location, Word len)
{
  // Parametry budou:
  //	-segment a offset, na nemz jsou data
  //	-linearni offset v ramci obrazovky (videosegmentu-B800h)
  //	-delka kopirovane oblasti - zadne vyrezy !!!

asm {
		push	ax
		push	cx
		push	ds
		push	es
		push	si
		push	di
		mov	si, src_off		// Si = From
		mov 	ax, VID_SEG
		mov	es, ax
		mov	di, location		// ES:DI = To
		mov	cx, len			// CX = How Many
		mov	ax, src_seg		// DS setting as last
		mov	ds, ax			//   because of DS change
		pushf
		cld
		rep	movsb			// Transfer
		popf
		pop	di
		pop	si
		pop	es
		pop	ds
		pop	cx
		pop	ax
  }
}

