/****************************************************************************/
/*                         TUOX kernel: modul PROC.C                        */
/****************************************************************************/
/* This file contains essentially all of the process and message handling.  */
/* It has two main entry points from the outside:                           */
/*                                                                          */
/*  sys_call:   called when a process or task does SEND, RECEIVE or SENDREC */
/*  k_interrupt: called by interrupt routines to send a message to task     */
/*                                                                          */
/* It also has five minor entry points:                                     */
/*                                                                          */
/*  ready:     put a process on one of the ready queues so it can be run    */
/*  unready:   remove a process from the ready queues                       */
/*  sched:     a process has run too long; schedule another one             */
/*  mini_send: send a message (used by interrupt signals, etc.)             */
/*  pick_proc: pick a process to run (used by system initialization)        */
/****************************************************************************/


#include "COMMON\type.h"
#include "COMMON\const.h"

#define EXTERN PUBLIC
#include "COMMON\glo.h"
#include "kernel\proc.h"
#include "test.cnf"
#include "test\kprintf.h"

#include "test\trap.h"

extern Word k_code_segment,k_data_segment;

/* --------------------- Message Tracing Code ----------------------------- */

#if (TEST_PHASE !=0 && defined (TEST_TRACE_MSG_PASSING))
 static Byte msgTraceLine = TEST_MSG_TRACE_1ST_LINE;
 static int msgNo=0;

 static void printMsgTrace(int function, int caller, int src_dest, message *m_ptr)
  {
   static message debug_mess;

   if (function & SEND) {

    if (caller==HARDWARE)
	 k_cp_mess(caller,
	  (DWord) ( (((DWord) k_data_segment) << CLICK_SHIFT) + (Word) (m_ptr)),
	  (DWord) ( (((DWord) k_data_segment) << CLICK_SHIFT) + (Word) (&debug_mess)));
     else
      k_cp_mess(caller,
	  (DWord) ( ( ((DWord) (proc[caller].p_map[D].mem_phys)) << CLICK_SHIFT)
			  +(Word) m_ptr),
	  (DWord) ( (((DWord) k_data_segment) << CLICK_SHIFT) + (Word) (&debug_mess)));
   }

   if (function==SEND)
    if (debug_mess.m_type!=TASK_REPLY)
     kprintf(msgTraceLine,0,7,"%i. S: %i -> %i, m_type=%i.",msgNo,caller,src_dest,debug_mess.m_type);
    else kprintf(msgTraceLine,0,7,"%i. S: %i -> %i, m_type=TASK_REPLY, rep_status=%i.",msgNo,caller,src_dest,debug_mess.REP_STATUS);
   else if (function==RECEIVE)
	   kprintf(msgTraceLine,0,7,"%i. R: %i <- %i.",msgNo,caller,src_dest);
      else
	kprintf(msgTraceLine,0,7,"%i. SR: %i <-> %i, m_type=%i.",msgNo,caller,src_dest,debug_mess.m_type);

   msgTraceLine++; msgNo++;
   if (msgTraceLine>=TEST_MSG_TRACE_LAST_LINE)
    msgTraceLine = TEST_MSG_TRACE_1ST_LINE;
  }
#endif
/* ------------------------------------------------------------------------ */

/*==========================================================================*/
/*                             k_interrupt                                  */
/*==========================================================================*/
PUBLIC void k_interrupt(int task_nr, message* m_ptr)
/*    task_nr ......  number of task to be started                          */
/*    m_ptr ......  interrupt message to send to the task                   */
{
  int i,j,k;    /* pouzito pro docasnou potrebu */
   /* An interrupt has occurred.  Schedule the task that handles it.
      Tato rutina je volana z nejnizsich assemblerovych vrstev.     */

   /* register struct proc *rp;    pointer to task's proc entry */
   /* rp = proc_addr(task_nr);  */

    /* Zkus zaslat message tasku pres mini_send() */
  if (mini_send(m_ptr->m_source, task_nr, m_ptr) != OK)
  /* pokud se nepovede: nastav v busy_map a uschovej message v task_mess[].
			Jde-li o clock, lost_click++ */
  { /* if */
    if (task_nr == CLOCK) lost_ticks++;
    busy_map |= TasknrToTaskbit(task_nr);
    task_mess[task_nr] = *m_ptr;
  } /* if */
  /* jinak: (zrus bit v busy map, kdyby byl nastaveny ) */
  else { busy_map &= (~TasknrToTaskbit(task_nr));
	 #if (TEST_PHASE !=0 && defined (TEST_TRACE_MSG_PASSING))
	  printMsgTrace(SEND,HARDWARE,task_nr,m_ptr);
	 #endif
       }

  /* zjisti, jestli bylo pri vstupu do rutiny neco v busy_map.
     Pokud ano: projed jednotlive bity busy_map a zkus zaslat
		message prislusnym taskum                     */

  if (busy_map)
  { /* if */
    for( i=1; i<(POC_BITU_PRER); i<<=1 )
    /* souvisi s TasknrToTaskbit() a TaskbitToTasknr() */
    { /* for */
      if( (j=(busy_map & i))!=0 )
      { /* if */
	k=TaskbitToTasknr(j);
	if (mini_send(task_mess[k].m_source,k,&(task_mess[k])) == OK) {
	  busy_map &= (~j);
	  #if (TEST_PHASE !=0 && defined (TEST_TRACE_MSG_PASSING))
	  printMsgTrace(SEND,HARDWARE,k,&(task_mess[k]));
	  #endif
	}
      } /* if */
    } /* for */
  } /* if */

  /* Pokud byl naplanovan task (v dusledku message) a bezi user proces,
     preplanuj */
  if ( (cur_proc >= NR_TASKS+NR_SERVS) || cur_proc==IDLE ) pick_proc();

  return;
} /* k_interrupt */



/*==========================================================================*/
/*                             pick_proc                                    */
/*==========================================================================*/
PUBLIC void pick_proc(void)
{
 /* Decide who to run now.*/

 /* Vyber nejprioritnejsi frontu, ve ktere je nekdo pripraven ke behu */
 int q;
 for(q=TASK_Q; q<NQ; q++)
   if (rdy_head[q] != NIL_PROC)
   {/*if*/
     prev_ptr=proc_ptr;
     cur_proc = rdy_head[q] - proc;
     proc_ptr = rdy_head[q];
     if (cur_proc >= NR_TASKS+NR_SERVS) bill_ptr = proc_ptr;
     return;
   }/*if*/
  cur_proc = IDLE;
  bill_ptr = proc_ptr = proc_addr(IDLE);
} /* pick_proc */



/*==========================================================================*/
/*                             sched                                        */
/*==========================================================================*/
PUBLIC void sched(void)
{
/* The current process has run too long. If another low priority (user)     */
/* process is runnable,put the current process on the end of the user queue,*/
/* possibly promoting another user to head of the queue.                    */


  /* disable interrupts */
  k_lock();

  /* pokud je USER_Q prazdna, nedelej nic */
  if (rdy_head[USER_Q] == NIL_PROC) return;

  /* vezmi proces z cele USER_Q a presun jej na konec */
  rdy_tail[USER_Q]->p_nextready = rdy_head[USER_Q];
  rdy_tail[USER_Q] = rdy_head[USER_Q];
  rdy_head[USER_Q] = rdy_head[USER_Q]->p_nextready;
  rdy_tail[USER_Q]->p_nextready = NIL_PROC;

  pick_proc();

  /* restore interrupts */
  k_restore();
  return;
} /* sched */


/* POTENCIALNI MODIFIKACE ZADA SOBE ZMENY MAKRA POC_BITU_PRER v PROC.H */
/*=====================================================================*/
/* TasknrToTaskbit je takova vec, ktera z cisla tasku vypocita         */
/* bit do busymap                                                      */
/*=====================================================================*/
PRIVATE unsigned TasknrToTaskbit(int task_nr)
{
  switch( task_nr )
  {
    case IDLE    : return 1;
    case CLOCK   : return 2;
    case FLOPPY  : return 4;
    case SYSTASK : return 8;
    case MEM     : return 16;
    case TTY     : return 32;
    case HDD     : return 64;
  } /* switch */
  return 0; /* nejjednodussi zpusob jak precurat prekladac bez pragmy */
} /* TasknrToTaskbit */


/*=====================================================================*/
/* TaskbitToTasknr je takova vec, ktera z bitu do busymap vypocita     */
/* cislo tasku                                                         */
/*=====================================================================*/
PRIVATE int TaskbitToTasknr(unsigned task_bit)
{
  switch( task_bit )
  {
    case  1: return IDLE;
    case  2: return CLOCK;
    case  4: return FLOPPY;
    case  8: return SYSTASK;
    case 16: return MEM;
    case 32: return TTY;
    case 64: return HDD;
  }
  return -1;   /* To je proste falesna hodnota, ktera nema co nastat. */
}


/*==========================================================================*/
/*                              sys_call                                    */
/*==========================================================================*/

PUBLIC int sys_call(int function, int caller, int src_dest, message *m_ptr)
/*int function;             SEND, RECEIVE, or BOTH
  int caller;               who is making this call
  int src_dest;             source to receive from or dest to send to
  message *m_ptr;           pointer to message */
{
/* The only system calls that exist in TUOX are sending and receiving
 * messages.  These are done by trapping to the kernel with
 * an INT instruction.
 * The trap is caught and sys_call() is called ( from s_call) to
 * send or receive a message (or both).
 */

/*
 * over platnost src_dest, a kdo muze s kym komunikovat. User_procesy
 * jen fn=BOTH, kdyz se to nepovede, nastav error registr (nejspis AX)
 * v preg[] procesu a nebav se s nim.
 */

  register struct proc *rp;
  int n;

  if (!isoksrc_dest(src_dest) && src_dest!=HARDWARE)
  {
    trap(TRAP_K_BADSRCDEST);
    return(E_BAD_SRC);
  }
  rp = proc_ptr;

  if (isuserp(rp) && (function != BOTH))
  {
    trap(TRAP_K_USERONLYBOTH);
    return(E_NO_PERM);
  }

  #if (TEST_PHASE !=0 && defined (TEST_TRACE_MSG_PASSING))
   printMsgTrace(function,caller,src_dest,m_ptr);
  #endif

  /* if (function & SEND) mini_send(caller, src_dest, m_ptr); */
  if (function & SEND)
  {
    /* Function = SEND or BOTH. */
    n = mini_send(caller, src_dest, m_ptr);
    if ((function == SEND) || (n != OK)) return(n); /* done, or SEND failed */
  }
  /* if (function & RECEIVE) mini_rec(caller, src_dest, m_ptr); */
  /* Function = RECEIVE or BOTH.                                */
  /* We have checked user calls are BOTH, and trust 'function' otherwise.*/

  return(mini_rec(caller, src_dest, m_ptr));
  /* Vzdy test jak to dopadlo a pripadny exit s nastavenim error registru */
} /* sys_call */


/*==========================================================================*/
/*                               mini_send                                  */
/*==========================================================================*/
PUBLIC int mini_send(int caller, int dest, message *m_ptr)
/* register struct proc *caller_ptr;  who is trying to send a message?      */
/*  int dest;                          to whom is message being sent?       */
/*  message *m_ptr;                    pointer to message buffer            */
{
/* Send a message from 'caller' to 'dest'. If 'dest' is blocked waiting for */
/* this message,copy the message to it and unblock 'dest'. If 'dest' is not */
/* waiting at all, or is waiting for another source, queue 'caller'.        */

/* caller muze byt HARDWARE (-1)                                            */
/* - pak to neni index do proc, ale message se musi vzit z int_mess,	    */
/*   resp. task_mess[]        						    */

  register struct proc *caller_ptr, *dest_ptr, *next_ptr;
  vir_bytes vb;                /* message buffer pointer as vir_bytes       */
  vir_clicks vlo, vhi;         /* virtual clicks containing message to send */

  /* Necht dest_ptr je pointer na slot proc[] ciloveho procesu,  */
  /* caller_ptr zdrojoveho                                       */

  dest_ptr = proc_addr(dest);       /* pointer to destination's proc entry */
  if (caller==HARDWARE) caller_ptr=proc_addr(IDLE);
   // jen aby tam neco bylo, aby prosly testy
   else caller_ptr = proc_addr(caller);

  /* User processes are only allowed to send to FS and MM. Check for this. */
  if (isuserp(caller_ptr) && !isservp(dest_ptr))
  { // funguje i pro caller=HARDWARE, protoze tehdy je caller_ptr=IDLE
    trap(TRAP_K_USERONLYTOSERV);
    return(E_BAD_DEST);
  }

  /* Test for dead destination process */
  if (dest_ptr->p_flags & P_SLOT_FREE)
  {
    trap(TRAP_K_DEADDEST);                      /* !!! tady konstanta od testeru !!! */
    return(E_BAD_DEST);
  }

  /* Check for deadlock by 'caller' and 'dest' sending to each other. */
  /* (shouldn't happenu if the system is written correctly */
  /* Do not test if caller=HARDWARE to save time
     (it would make no sense anyway) */

  if ((dest_ptr->p_flags & SENDING) && (caller!=HARDWARE))
  { /* if */
    next_ptr = dest_ptr->p_sendlink;
    while (TRUE)
    { /* while */
      if (next_ptr == caller_ptr) return(E_LOCKED);
      if (next_ptr->p_flags & SENDING)
      {
	next_ptr = next_ptr->p_sendlink;
      }
      else
	break;
    } /* while */
  } /* if */

  /* Check to see if 'dest' is blocked waiting for this
     message (from ANY or me). */

  if ( (dest_ptr->p_flags & RECEIVING) == RECEIVING &&
       (dest_ptr->p_getfrom == ANY || dest_ptr->p_getfrom == caller) )
  {
    if (caller==HARDWARE)
     /* P.G.
      message muze byt ulozeno budto v int_mess (pokud to proslo primo
      pres obsluznou rutinu preruseni (mpx88.c) a k_interrupt,
      nebo v task_mess[], pokud tam bylo uschovano. V kazdem pripade
      ale m_ptr ukazuje NA MESSAGE, jelikoz jak int_mess tak task_mess
      jsou soucasti adr. prostoru jadra. Proto se zdroj pro kopirovani
      bere z adresy m_ptr.
     */
	 k_cp_mess(caller,
	  (DWord) ( (((DWord) k_data_segment) << CLICK_SHIFT) + (Word) m_ptr),
	  (DWord) ( ( ((DWord) (proc[dest].p_map[D].mem_phys)) << CLICK_SHIFT)
			  +(Word) (proc[dest].p_messbuf)));
     else
      k_cp_mess(caller,
	    (DWord) ( ( ((DWord) (proc[caller].p_map[D].mem_phys)) << CLICK_SHIFT)
			  +(Word) m_ptr),
	    (DWord) ( ( ((DWord) (proc[dest].p_map[D].mem_phys)) << CLICK_SHIFT)
			  +(Word) (proc[dest].p_messbuf)) );

    dest_ptr->p_flags &= ~RECEIVING;        /* deblock destination */
    if (dest_ptr->p_flags == 0) ready(dest_ptr);
  }
  else
  { /* Destination is not waiting.                         */
    /* Zablokuj volajici proces a pridej do fronty procesu */
    /* cekajicich na vyslani zpravy ciloveho procesu       */

    if (caller == HARDWARE) return(E_OVERRUN);

    caller_ptr->p_messbuf = m_ptr;
    if (caller_ptr->p_flags == 0) unready(caller_ptr);
    caller_ptr->p_flags |= SENDING;
     /* Process is now blocked.  Put in on the destination's queue. */

    if ( (next_ptr = dest_ptr->p_callerq) == NIL_PROC)
    {
      dest_ptr->p_callerq = caller_ptr;
    }
    else
    {
      while (next_ptr->p_sendlink != NIL_PROC)
	next_ptr = next_ptr->p_sendlink;
	next_ptr->p_sendlink = caller_ptr;
    }
    caller_ptr->p_sendlink = NIL_PROC;
  }

  return(OK);
} /* mini_send */


/*==========================================================================*/
/*                             mini_rec                                     */
/*==========================================================================*/
PRIVATE int mini_rec(int caller, int src, message *m_ptr)
{
  register struct proc *caller_ptr; /* process trying to get message        */
  register struct proc *src_ptr;    /* process trying to send message to us */
  /* int src;                       which message source is wanted (or ANY) */
  /* message *m_ptr;                  pointer to message buffer             */

/* A process or task wants to get a message.  If one is already queued,     */
/* acquire it and deblock the sender. If no message from the desired source */
/* is available, block the caller. No need to check parameters for validity.*/
/* Users calls are always sendrec(), and mini_send() has checked already.   */
/* Calls from the tasks, MM, and FS are trusted.                            */

  register struct proc *pptr,*oldptr;
  procnr_t realsrc;

  caller_ptr=proc_addr(caller);
  pptr=proc[caller].p_callerq;

  if (src==ANY) {
   src_ptr=pptr; realsrc=pptr-proc;
   if (src_ptr!=NULL) proc[caller].p_callerq=proc[caller].p_callerq->p_sendlink;
  }
   else {
    src_ptr=oldptr=NULL;

    while (pptr!=NULL) {
     if ((pptr-proc)==src)  { realsrc=pptr-proc; src_ptr=pptr; break; }
     oldptr=pptr;
     pptr=pptr->p_sendlink;
    }

   if(src_ptr!=NULL) oldptr->p_sendlink=src_ptr->p_sendlink;
   }



  /* test if a message from desired source is already available. */

  if (src_ptr!=NULL)
  {
    k_cp_mess(realsrc,
	     (DWord)( (((DWord) (proc[realsrc].p_map[D].mem_phys)) << CLICK_SHIFT)
			  +(Word) (src_ptr->p_messbuf)),
	     (DWord)( (((DWord) (proc[caller].p_map[D].mem_phys))<<CLICK_SHIFT)
			  +(Word) m_ptr));

    src_ptr->p_flags &= ~SENDING;
    if (src_ptr->p_flags == 0) ready(src_ptr);
  }
  else
  {
    /* No suitable message is available.         */
    /* Block the process trying to receive.      */
    /* nastav proc[].p_getfrom, proc[].p_messbuf */
    caller_ptr->p_getfrom = src;
    caller_ptr->p_messbuf = m_ptr;
    if (caller_ptr->p_flags == 0) unready(caller_ptr);
    caller_ptr->p_flags |= RECEIVING;
  }
  return(OK);
} /* mini_rec */


/*==========================================================================*/
/*                             ready                                        */
/*==========================================================================*/
PUBLIC void ready(register struct proc *rp)
       /* this process is now runnable */
{
  /* Add 'rp' to the end of one of the queues of runnable processes. */

  /* disable interrupts */
  k_lock();
  /* podle pozice procesu v proc[] rozhodni o fronte, pripoj */
  if (istaskp(rp))
  {
    if (rdy_head[TASK_Q] != NIL_PROC) rdy_tail[TASK_Q]->p_nextready = rp;
      /* Add to tail of nonempty queue. */
    else rdy_head[TASK_Q] = rp;
      /* add to empty queue */

     rdy_tail[TASK_Q] = rp;
     rp->p_nextready = NIL_PROC;     /* new entry has no successor */

     /* restore interrupts */
    k_restore();
    return;
  }

  /* others are similar */
  if (!isuserp(rp))
  {
    if (rdy_head[SERVER_Q] != NIL_PROC) rdy_tail[SERVER_Q]->p_nextready = rp;
    else rdy_head[SERVER_Q] = rp;

    rdy_tail[SERVER_Q] = rp;
    rp->p_nextready = NIL_PROC;

    /* restore interrupts */
    k_restore();
    return;
  }

  if (rdy_head[USER_Q] != NIL_PROC) rdy_tail[USER_Q]->p_nextready = rp;
  else rdy_head[USER_Q] = rp;
  rdy_tail[USER_Q] = rp;
  rp->p_nextready = NIL_PROC;

  /* restore interrupts */
  k_restore();
} /* ready */


/*==========================================================================*/
/*                             unready                                      */
/*==========================================================================*/
PUBLIC void unready(register struct proc *rp)
       /* this process is no longer runnable */
{
/*podle pozice procesu v proc[] rozhodni o fronte, najdi v ni proces a vyhod*/
  register struct proc *xp;

  register int q;
/*  
Pozn.:
  tasky a servery: odstranuje se pouze proces z cela fronty
  (byl aktivni a ted se dobrovolne vzdal rizeni pre SEND|RECV)
  U procesu: muze se chtit odstranit kterykoliv proces, ac neni prave
  bezici (napr. prisel signal KILL na nej)
*/

  /* disable interrupts */
  k_lock();
  q= (istaskp(rp) ? TASK_Q : (isservp(rp) ? SERVER_Q : USER_Q));

  if ( (xp = rdy_head[q]) == NIL_PROC)
    {  k_restore(); return; }

  if (xp == rp)
    {
      /* Remove head of queue */
      rdy_head[q] = xp->p_nextready;

      /* Kdyz byl ve fronte prave jeden => Upravit ocas ! */
      if (rdy_head[q]==NULL) rdy_tail[q]=NULL;

      if (rp == proc_ptr) pick_proc();

      /* restore interrupts */
      k_restore();
      return;
    }
  else {
  /* Search body of queue.  A process can be made unready even if it is  */
  /* not running by being sent a signal that kills it.                   */

  while (xp->p_nextready != rp)
    if ( (xp = xp->p_nextready) == NIL_PROC)
     { k_restore(); return; /* not found in the queue */ }
  }

  xp->p_nextready = xp->p_nextready->p_nextready; // detach from queue */

  while (xp->p_nextready!=NIL_PROC) xp=xp->p_nextready;
  rdy_tail[q]=xp;

  /* restore interrupts */
  k_restore();
  } /* unready */
